#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#include <SD.h>

// motor pins
#define ENA 4
#define IN1 5
#define IN2 6
#define IN3 7
#define IN4 8
#define ENB 9

// speed sensor pins
#define SPEED_SENSOR_1 2
#define SPEED_SENSOR_2 3

// SD card pin
#define SD_CS_PIN 53

// LCD setup
LiquidCrystal_I2C lcd(0x27, 16, 2);

// variables for random movement and RPM calculation
long lastMoveTime = 0;
int action = 0;
volatile unsigned long pulseCount1 = 0;
volatile unsigned long pulseCount2 = 0;
unsigned long rpm1 = 0;
unsigned long rpm2 = 0;

void setup() {
  // serial for debugging
  Serial.begin(9600);

  // Motor pins
  pinMode(ENA, OUTPUT);
  pinMode(IN1, OUTPUT);
  pinMode(IN2, OUTPUT);
  pinMode(IN3, OUTPUT);
  pinMode(IN4, OUTPUT);
  pinMode(ENB, OUTPUT);

  // Speed sensor pins
  pinMode(SPEED_SENSOR_1, INPUT);
  pinMode(SPEED_SENSOR_2, INPUT);

  // Attach interrupts to the speed sensors
  attachInterrupt(digitalPinToInterrupt(SPEED_SENSOR_1), countPulse1, RISING);
  attachInterrupt(digitalPinToInterrupt(SPEED_SENSOR_2), countPulse2, RISING);

  // LCD initialization
  lcd.init();
  lcd.backlight();

  // set initial motor speed
  analogWrite(ENA, 255); // Full speed
  analogWrite(ENB, 255); // Full speed

  // random seed
  randomSeed(analogRead(0));

  // SD card initialization
  if (!SD.begin(SD_CS_PIN)) {
    Serial.println("Initialization failed!");
    lcd.setCursor(0, 0);
    lcd.print("SD init fail");
    return;
  }
  Serial.println("Initialization done.");
}

void loop() {
  // check if 1 second has passed
  if (millis() - lastMoveTime > 1000) {
    lastMoveTime = millis();
    
    // calculate RPM for each sensor
    rpm1 = (pulseCount1 * 60) / 20;
    rpm2 = (pulseCount2 * 60) / 20; // Assuming 20 pulses per revolution
    
    // reset pulse counts
    pulseCount1 = 0;
    pulseCount2 = 0;

    // display RPM on the LCD
    lcd.setCursor(0, 0);
    lcd.print("RPM1:");
    lcd.print(rpm1);
    lcd.setCursor(0, 1);
    lcd.print("RPM2:");
    lcd.print(rpm2);

    // debugging output
    Serial.print("RPM1: ");
    Serial.print(rpm1);
    Serial.print(" RPM2: ");
    Serial.println(rpm2);
    
    // write RPM data to SD card
    writeRPMToSD(rpm1, rpm2);
    
    // 0: forward, 1: backward, 2: left, 3: right
    action = random(4);
    
    switch(action) {
      case 0:
        moveForward();
        break;
      case 1:
        moveBackward();
        break;
      case 2:
        turnLeft();
        break;
      case 3:
        turnRight();
        break;
    }
  }
}

void countPulse1() {
  pulseCount1++;
}

void countPulse2() {
  pulseCount2++;
}

void moveForward() {
  digitalWrite(IN1, HIGH);
  digitalWrite(IN2, LOW);
  digitalWrite(IN3, HIGH);
  digitalWrite(IN4, LOW);
}

void moveBackward() {
  digitalWrite(IN1, LOW);
  digitalWrite(IN2, HIGH);
  digitalWrite(IN3, LOW);
  digitalWrite(IN4, HIGH);
}

void turnLeft() {
  digitalWrite(IN1, LOW);
  digitalWrite(IN2, HIGH);
  digitalWrite(IN3, HIGH);
  digitalWrite(IN4, LOW);
}

void turnRight() {
  digitalWrite(IN1, HIGH);
  digitalWrite(IN2, LOW);
  digitalWrite(IN3, LOW);
  digitalWrite(IN4, HIGH);
}

void writeRPMToSD(unsigned long rpm1, unsigned long rpm2) {
  File dataFile = SD.open("rpm_data.txt", FILE_WRITE);
  
  if (dataFile) {
    dataFile.print("RPM1: ");
    dataFile.print(rpm1);
    dataFile.print(" RPM2: ");
    dataFile.println(rpm2);
    dataFile.close();
    Serial.println("RPM data written to SD card.");
  }
}